@@ -15,9 +15,13 @@ const _ = Gettext.gettext;
 const WORKSPACE_SCHEMA = 'org.gnome.desktop.wm.preferences';
 const WORKSPACE_KEY = 'workspace-names';
 
-let WindowPreview = GObject.registerClass({
-    GTypeName: 'WorkspaceIndicatorWindowPreview',
-}, class WindowPreview extends St.Button {
+const TOOLTIP_OFFSET = 6;
+const TOOLTIP_ANIMATION_TIME = 150;
+
+const MAX_THUMBNAILS = 6;
+
+let WindowPreview = GObject.registerClass(
+class WindowPreview extends St.Button {
     _init(window) {
         super._init({
             style_class: 'workspace-indicator-window-preview',
@@ -31,27 +35,14 @@ let WindowPreview = GObject.registerClass({
         this.connect('destroy', this._onDestroy.bind(this));
 
         this._sizeChangedId = this._window.connect('size-changed',
-            this._relayout.bind(this));
+            () => this.queue_relayout());
         this._positionChangedId = this._window.connect('position-changed',
-            this._relayout.bind(this));
-        this._minimizedChangedId = this._window.connect('notify::minimized',
-            this._relayout.bind(this));
-        this._monitorEnteredId = global.display.connect('window-entered-monitor',
-            this._relayout.bind(this));
-        this._monitorLeftId = global.display.connect('window-left-monitor',
-            this._relayout.bind(this));
-
-        // Do initial layout when we get a parent
-        let id = this.connect('parent-set', () => {
-            this.disconnect(id);
-            if (!this.get_parent())
-                return;
-            this._laterId = Meta.later_add(Meta.LaterType.BEFORE_REDRAW, () => {
-                this._laterId = 0;
-                this._relayout();
-                return false;
+            () => {
+                this._updateVisible();
+                this.queue_relayout();
             });
-        });
+        this._minimizedChangedId = this._window.connect('notify::minimized',
+            this._updateVisible.bind(this));
 
         this._focusChangedId = global.display.connect('notify::focus-window',
             this._onFocusChanged.bind(this));
@@ -59,19 +50,15 @@ let WindowPreview = GObject.registerClass({
     }
 
     // needed for DND
-    get realWindow() {
-        return this._window.get_compositor_private();
+    get metaWindow() {
+        return this._window;
     }
 
     _onDestroy() {
         this._window.disconnect(this._sizeChangedId);
         this._window.disconnect(this._positionChangedId);
         this._window.disconnect(this._minimizedChangedId);
-        global.display.disconnect(this._monitorEnteredId);
-        global.display.disconnect(this._monitorLeftId);
         global.display.disconnect(this._focusChangedId);
-        if (this._laterId)
-            Meta.later_remove(this._laterId);
     }
 
     _onFocusChanged() {
@@ -81,44 +68,64 @@ let WindowPreview = GObject.registerClass({
             this.remove_style_class_name('active');
     }
 
-    _relayout() {
-        let monitor = Main.layoutManager.findIndexForActor(this);
-        this.visible = monitor === this._window.get_monitor() &&
+    _updateVisible() {
+        const monitor = Main.layoutManager.findIndexForActor(this);
+        const workArea = Main.layoutManager.getWorkAreaForMonitor(monitor);
+        this.visible = this._window.get_frame_rect().overlap(workArea) &&
             this._window.window_type !== Meta.WindowType.DESKTOP &&
             this._window.showing_on_its_workspace();
-
-        if (!this.visible)
-            return;
-
-        let workArea = Main.layoutManager.getWorkAreaForMonitor(monitor);
-        let hscale = this.get_parent().allocation.get_width() / workArea.width;
-        let vscale = this.get_parent().allocation.get_height() / workArea.height;
-
-        let frameRect = this._window.get_frame_rect();
-        this.set_size(
-            Math.round(Math.min(frameRect.width, workArea.width) * hscale),
-            Math.round(Math.min(frameRect.height, workArea.height) * vscale));
-        this.set_position(
-            Math.round(frameRect.x * hscale),
-            Math.round(frameRect.y * vscale));
     }
 });
 
-let WorkspaceThumbnail = GObject.registerClass({
-    GTypeName: 'WorkspaceIndicatorWorkspaceThumbnail',
-}, class WorkspaceThumbnail extends St.Button {
+let WorkspaceLayout = GObject.registerClass(
+class WorkspaceLayout extends Clutter.LayoutManager {
+    vfunc_get_preferred_width() {
+        return [0, 0];
+    }
+
+    vfunc_get_preferred_height() {
+        return [0, 0];
+    }
+
+    vfunc_allocate(container, box) {
+        const monitor = Main.layoutManager.findIndexForActor(container);
+        const workArea = Main.layoutManager.getWorkAreaForMonitor(monitor);
+        const hscale = box.get_width() / workArea.width;
+        const vscale = box.get_height() / workArea.height;
+
+        for (const child of container) {
+            const childBox = new Clutter.ActorBox();
+            const frameRect = child.metaWindow.get_frame_rect();
+            childBox.set_size(
+                Math.round(Math.min(frameRect.width, workArea.width) * hscale),
+                Math.round(Math.min(frameRect.height, workArea.height) * vscale));
+            childBox.set_origin(
+                Math.round((frameRect.x - workArea.x) * hscale),
+                Math.round((frameRect.y - workArea.y) * vscale));
+            child.allocate(childBox);
+        }
+    }
+});
+
+let WorkspaceThumbnail = GObject.registerClass(
+class WorkspaceThumbnail extends St.Button {
     _init(index) {
         super._init({
             style_class: 'workspace',
             child: new Clutter.Actor({
-                layout_manager: new Clutter.BinLayout(),
+                layout_manager: new WorkspaceLayout(),
                 clip_to_allocation: true,
             }),
-            x_fill: true,
-            y_fill: true,
         });
 
+        this._tooltip = new St.Label({
+            style_class: 'dash-label',
+            visible: false,
+        });
+        Main.uiGroup.add_child(this._tooltip);
+
         this.connect('destroy', this._onDestroy.bind(this));
+        this.connect('notify::hover', this._syncTooltip.bind(this));
 
         this._index = index;
         this._delegate = this; // needed for DND
@@ -144,16 +151,15 @@ let WorkspaceThumbnail = GObject.registerClass({
     }
 
     acceptDrop(source) {
-        if (!source.realWindow)
+        if (!source.metaWindow)
             return false;
 
-        let window = source.realWindow.get_meta_window();
-        this._moveWindow(window);
+        this._moveWindow(source.metaWindow);
         return true;
     }
 
     handleDragOver(source) {
-        if (source.realWindow)
+        if (source.metaWindow)
             return DND.DragMotionResult.MOVE_DROP;
         else
             return DND.DragMotionResult.CONTINUE;
@@ -204,7 +210,39 @@ let WorkspaceThumbnail = GObject.registerClass({
             ws.activate(global.get_current_time());
     }
 
+    _syncTooltip() {
+        if (this.hover) {
+            this._tooltip.set({
+                text: Meta.prefs_get_workspace_name(this._index),
+                visible: true,
+                opacity: 0,
+            });
+
+            const [stageX, stageY] = this.get_transformed_position();
+            const thumbWidth = this.allocation.get_width();
+            const thumbHeight = this.allocation.get_height();
+            const tipWidth = this._tooltip.width;
+            const xOffset = Math.floor((thumbWidth - tipWidth) / 2);
+            const monitor = Main.layoutManager.findMonitorForActor(this);
+            const x = Math.clamp(
+                stageX + xOffset,
+                monitor.x,
+                monitor.x + monitor.width - tipWidth);
+            const y = stageY + thumbHeight + TOOLTIP_OFFSET;
+            this._tooltip.set_position(x, y);
+        }
+
+        this._tooltip.ease({
+            opacity: this.hover ? 255 : 0,
+            duration: TOOLTIP_ANIMATION_TIME,
+            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
+            onComplete: () => (this._tooltip.visible = this.hover),
+        });
+    }
+
     _onDestroy() {
+        this._tooltip.destroy();
+
         this._workspace.disconnect(this._windowAddedId);
         this._workspace.disconnect(this._windowRemovedId);
         global.display.disconnect(this._restackedId);
@@ -252,14 +290,14 @@ class WorkspaceIndicator extends PanelMenu.Button {
             workspaceManager.connect_after('workspace-switched',
                 this._onWorkspaceSwitched.bind(this)),
             workspaceManager.connect('notify::layout-rows',
-                this._onWorkspaceOrientationChanged.bind(this)),
+                this._updateThumbnailVisibility.bind(this)),
         ];
 
         this.connect('scroll-event', this._onScrollEvent.bind(this));
         this._thumbnailsBox.connect('scroll-event', this._onScrollEvent.bind(this));
         this._createWorkspacesSection();
         this._updateThumbnails();
-        this._onWorkspaceOrientationChanged();
+        this._updateThumbnailVisibility();
 
         this._settings = new Gio.Settings({ schema_id: WORKSPACE_SCHEMA });
         this._settingsChangedId = this._settings.connect(
@@ -281,16 +319,19 @@ class WorkspaceIndicator extends PanelMenu.Button {
         super._onDestroy();
     }
 
-    _onWorkspaceOrientationChanged() {
-        let vertical = global.workspace_manager.layout_rows === -1;
-        this.reactive = vertical;
+    _updateThumbnailVisibility() {
+        const { workspaceManager } = global;
+        const vertical = workspaceManager.layout_rows === -1;
+        const useMenu =
+            vertical || workspaceManager.n_workspaces > MAX_THUMBNAILS;
+        this.reactive = useMenu;
 
-        this._statusLabel.visible = vertical;
-        this._thumbnailsBox.visible = !vertical;
+        this._statusLabel.visible = useMenu;
+        this._thumbnailsBox.visible = !useMenu;
 
         // Disable offscreen-redirect when showing the workspace switcher
         // so that clip-to-allocation works
-        Main.panel.set_offscreen_redirect(vertical
+        Main.panel.set_offscreen_redirect(useMenu
             ? Clutter.OffscreenRedirect.ALWAYS
             : Clutter.OffscreenRedirect.AUTOMATIC_FOR_OPACITY);
     }
@@ -307,6 +348,7 @@ class WorkspaceIndicator extends PanelMenu.Button {
     _nWorkspacesChanged() {
         this._createWorkspacesSection();
         this._updateThumbnails();
+        this._updateThumbnailVisibility();
     }
 
     _updateMenuOrnament() {
